import functools



class Filter(object):
    def __and__(self, o):
        assert isinstance(o, Filter)
        return AndFilter(self, o)

    def __or__(self, o):
        assert isinstance(o, Filter)
        return OrFilter(self, o)

    def _filter(self, trace):
        return True

class BinaryFilter(Filter):
    def __init__(self, left, right):
        self.left = left
        self.right = right

class AndFilter(BinaryFilter):
    def _filter(self, trace):
        return self.left._filter(trace) and self.right._filter(trace)

class OrFilter(BinaryFilter):
    def _filter(self, trace):
        if self.left._filter(trace):
            return True
        if self.right._filter(trace):
            return True
        return False

class OpFilter(Filter):
    def __init__(self, name):
        self.name = name

    def _filter(self, trace):
        name = self.name
        for stage in trace.stages.keys():
            for op in trace.get_stage(stage).get_ops():
                if name in op.opname:
                    return True
        return False

class FuncFilter(Filter):
    def __init__(self, name):
        self.name = name

    def _filter(self, trace):
        name = self.name
        stage = trace.get_stage('noopt')
        if not stage:
            return False

        for mp in stage.get_merge_points():
            if name in mp.get_scope():
                return True

        return False

class LoopFilter(Filter):
    def _filter(self, trace):
        return trace.type == 'loop'

class BridgeFilter(Filter):
    def _filter(self, trace):
        return trace.type == 'bridge'

loops = LoopFilter()

bridges = BridgeFilter()


QUERY_API = {
    # functions
    'op': OpFilter,
    'func': FuncFilter,
    # variable filters without state
    'loops': loops,
    'bridges': bridges,
}


class Query(object):
    def __init__(self, text):
        self.query_str = text
        self.forest = None

    def __call__(self, forest):
        self.forest = forest
        return self.evaluate(forest, self.query_str)

    def evaluate(self, forest, qstr):
        if qstr is None or qstr.strip() == '':
            return None

        api = {}
        for k,f in QUERY_API.items():
            api[k] = f
        # ---------------------------------------------
        # SECURITY ISSUE:
        # never execute this in the server environment
        # ---------------------------------------------
        f = eval(qstr, {}, api)
        return [t for t in forest.traces.values() if f._filter(t)]

def new_unsafe_query(query):
    return Query(query)





# coding: utf-8
import sys
import py
from jitlog.objects import (TraceForest, MergePoint, FlatOp)
from jitlog import constants as const

PY3 = sys.version_info[0] >= 3

@py.test.mark.parametrize('encoding,text,decoded,bom',
    [('ascii',   b'a!1%$', u'a!1%$', None),
     ('utf-8',   b"\x41\xE2\x89\xA2\xCE\x91\x2E", u'A≢Α.', None),
     ('latin-1', b'\xDCber', u'Über', None),
    ])
def test_merge_point_extract_source_code(encoding,text,decoded,bom,tmpdir):
    forest = TraceForest(1)
    trace = forest.add_trace('loop', 0, 0)
    trace.start_mark(const.MARK_TRACE_OPT)
    file = tmpdir.join("file"+encoding+".py")
    l = []
    if bom:
        l.append(bom)
    l.append("# coding: ".encode(encoding))
    l.append(encoding.encode(encoding))
    l.append("\r\n".encode(encoding))
    l.append("print(\"".encode(encoding))
    l.append(text)
    l.append("\")".encode(encoding))
    l.append("\r\n".encode(encoding))
    file.write_binary(b''.join(l))
    trace.add_instr(MergePoint({0x1: str(file), 0x2: 2}))
    forest.extract_source_code_lines()
    line = forest.source_lines[str(file)][2]
    if PY3:
        assert line == (0, "print(\"" + decoded + "\")")
    else:
        assert line == (0, "print(\"" + text + "\")")



import os
import sys
import gzip
import tempfile
import requests
try:
    from urlparse import urlparse
except ImportError:
    # python 3
    from urllib.parse import urlparse

def compress_file(filename):
    fileno, name = tempfile.mkstemp(prefix='jit', suffix='.log.zip')
    os.close(fileno)
    with open(filename, 'rb') as fd:
        with gzip.open(name, 'wb') as zipfd:
            while True:
                chunk = fd.read(1024)
                if not chunk:
                    break
                zipfd.write(chunk)
    return name

def upload(filepath, url):
    zfilepath = compress_file(filepath)
    with open(zfilepath, 'rb') as fd:
        r = requests.post(url, files={ 'file': fd })
        if r.status_code != 200:
            sys.stderr.write("PyPy JIT log: Server rejected file. status: %d, msg: '%s'\n" % \
                    (r.status_code,r.text))
            return
        checksum = r.text[1:-1]
        assert checksum != ""
        netloc = urlparse(url).netloc
        sys.stderr.write("PyPy JIT log: http://%s/#/%s/traces\n" % (netloc, checksum))




from __future__ import print_function

import argparse
import sys
from six.moves import configparser


def build_argparser():
    parser = argparse.ArgumentParser(
        description='VMprof',
        prog="vmprof"
    )
    parser.add_argument(
        'program',
        help='program'
    )

    parser.add_argument(
        'args',
        nargs=argparse.REMAINDER,
        help='program arguments'
    )

    parser.add_argument(
        '--config',
        type=argparse.FileType('r'),
    )

    parser.add_argument(
        '--period', '-p',
        type=float,
        default=0.001,
        help='Sampling period (in microseconds)'
    )

    parser.add_argument(
        '--web-auth',
        help='Authtoken for your acount on the server, works only when --web is used'
    )

    parser.add_argument(
        '--web-url',
        metavar='url',
        default='http://vmprof.com',
        help='Provide URL instead of the default vmprof.com)'
    )
    parser.add_argument(
        '--mem',
        action="store_true",
        help='Do memory profiling as well',
    )
    parser.add_argument(
        '--lines',
        action="store_true",
        help='Store lines execution stats',
    )
    parser.add_argument(
        '--jitlog',
        action='store_true',
        help='Upload the jitlog to remote server (defaults to vmprof.com)',
    )
    parser.add_argument(
        '--no-native', '-n',
        default=not sys.platform.startswith('linux'),
        action='store_true',
        help='Disable native profiling for this run'
    )
    output_mode_args = parser.add_mutually_exclusive_group()
    output_mode_args.add_argument(
        '--web',
        action='store_true',
        help='Upload profiling stats to a remote server (defaults to vmprof.com)'
    )
    output_mode_args.add_argument(
        '--output', '-o',
        metavar='file.prof',
        type=argparse.FileType('w+b'),
        help='Save profiling data to file'
    )

    return parser


def parse_args(argv):
    parser = build_argparser()
    args = parser.parse_args(argv)
    if args.config:
        ini_options = [
            ('period', float),
            ('web', str),
            ('mem', bool),
            ('web-auth', str),
            ('web-url', str),
            ('output', str),
            ('no-native', bool),
        ]

        ini_parser = IniParser(args.config)

        for name, type in ini_options:
            argname = name.replace("-", "_")
            default = parser.get_default(argname)
            current = getattr(args, argname, default)
            if current == default:
                value = ini_parser.get_option(name, type, default)
                setattr(args, argname, value)

    return args


class IniParser(object):

    def __init__(self, f):
        self.ini_parser = configparser.ConfigParser()
        self.ini_parser.readfp(f)

    def get_option(self, name, type, default=None):
        if type == float:
            try:
                return self.ini_parser.getfloat('global', name)
            except configparser.NoOptionError:
                return default
        elif type == bool:
            try:
                return self.ini_parser.getboolean('global', name)
            except configparser.NoOptionError:
                return default

        try:
            return self.ini_parser.get('global', name)
        except configparser.NoOptionError:
            return default


def show(stats):
    p = stats.top_profile()
    if not p:
        print("No stats, profile is empty!\nDid you program run long enough (more than 1 second)?")
        return

    p.sort(key=lambda x: x[1], reverse=True)
    top = p[0][1]

    max_len = max([_namelen(e[0]) for e in p])

    print(" vmprof output:")
    print(" %:      name:" + " " * (max_len - 3) + "location:")

    for k, v in p:
        v = "%.1f%%" % (float(v) / top * 100)
        if v == '0.0%':
            v = '<0.1%'
        if k.startswith('py:'):
            try:
                _, func_name, lineno, filename = k.split(":", 3)
                lineno = int(lineno)
            except ValueError:
                print(" %s %s" % (v.ljust(7), k.ljust(max_len + 1)))
                # badly done split
            else:
                print(" %s %s %s:%d" % (v.ljust(7), func_name.ljust(max_len + 1), filename, lineno))
        else:
            print(" %s %s" % (v.ljust(7), k.ljust(max_len + 1)))


def _namelen(e):
    if e.startswith('py:'):
        return len(e.split(':')[1])
    else:
        return len(e)



from jitlog.objects import (TraceForest, FlatOp, MergePoint)
from jitlog import constants as c
from jitlog import query

class TestQueries(object):
    def q(self, forest, text):
        return query.new_unsafe_query(text)(forest)

    def test_query_empty_forest(self):
        f = TraceForest(3, is_32bit=False, machine='s390x')
        assert self.q(f, '') == None
        assert self.q(f, 'loops & bridges') == []

    def test_query_small_forest(self):
        f = TraceForest(3, is_32bit=False, machine='s390x')
        #
        t = f.add_trace('loop', 0, 0, 'jd')
        stage = t.start_mark(c.MARK_TRACE_OPT)
        stage.append_op(FlatOp(0, 'load', [], '?'))
        #
        t2 = f.add_trace('loop', 1, 1, 'jd')
        stage = t2.start_mark(c.MARK_TRACE_OPT)
        stage.append_op(FlatOp(0, 'store', [], '?'))
        #
        assert len(f.traces) == 2
        assert self.q(f, '') == None
        assert self.q(f, 'op("load")') == [t]

    def test_query_loops_and_bridges(self):
        f = TraceForest(3, is_32bit=False, machine='s390x')
        #
        t = f.add_trace('loop', 0, 0, 'jd')
        t2 = f.add_trace('bridge', 1, 1, 'jd')
        #
        assert len(f.traces) == 2
        assert self.q(f, 'loops') == [t]
        assert self.q(f, 'bridges') == [t2]

    def test_filter(self):
        from jitlog.query import loops, bridges, Filter
        f = TraceForest(3, is_32bit=False, machine='s390x')
        loop = f.add_trace('loop', 0, 0, 'su')
        bridge = f.add_trace('bridge', 1, 1, 'shi')
        assert loops._filter(loop)
        assert not loops._filter(bridge)
        assert not bridges._filter(loop)
        assert bridges._filter(bridge)
        r = loops | bridges
        assert isinstance(r, Filter)
        assert r._filter(loop)
        assert r._filter(bridge)

    def test_func_filter(self):
        from jitlog.query import FuncFilter as func
        f = TraceForest(3, is_32bit=False, machine='s390x')
        loop = f.add_trace('loop', 0, 0, 'su')
        stage = loop.start_mark(c.MARK_TRACE)
        stage.append_op(MergePoint({ c.MP_SCOPE[0]: '_sake_in_a_glass' }))
        assert not func('hello')._filter(loop)
        assert func('sake')._filter(loop)
        assert not func('s_a_k_e')._filter(loop)




import os
import sys
try:
    from shutil import which
except ImportError:
    from backports.shutil_which import which

import _vmprof

from vmprof import cli

from vmprof.reader import (MARKER_NATIVE_SYMBOLS, FdWrapper,
        LogReaderState, LogReaderDumpNative)
from vmprof.stats import Stats
from vmprof.profiler import Profiler, read_profile


PY3  = sys.version_info[0] >= 3
IS_PYPY = '__pypy__' in sys.builtin_module_names

# it's not a good idea to use a "round" default sampling period, else we risk
# to oversample periodic tasks which happens to run at e.g. 100Hz or 1000Hz:
# http://www.solarisinternals.com/wiki/index.php/DTrace_Topics_Hints_Tips#profile-1001.2C_profile-997.3F
#
# To avoid the problem, we use a period which is "almost" but not exactly
# 1000Hz
DEFAULT_PERIOD = 0.00099

def disable():
    try:
        # fish the file descriptor that is still open!
        if hasattr(_vmprof, 'stop_sampling'):
            fileno = _vmprof.stop_sampling()
            if fileno >= 0:
                # TODO does fileobj leak the fd? I dont think so, but need to check 
                fileobj = FdWrapper(fileno)
                l = LogReaderDumpNative(fileobj, LogReaderState())
                l.read_all()
                if hasattr(_vmprof, 'write_all_code_objects'):
                    _vmprof.write_all_code_objects(l.dedup)
        _vmprof.disable()
    except IOError as e:
        raise Exception("Error while writing profile: " + str(e))

def _is_native_enabled(native):
    if os.name == "nt":
        if native:
            raise ValueError("native profiling is only supported on Linux & Mac OS X")
        native = False
    else:
        if native is None:
            native = True
    return native

if IS_PYPY:
    def enable(fileno, period=DEFAULT_PERIOD, memory=False, lines=False, native=None, real_time=False, warn=True):
        pypy_version_info = sys.pypy_version_info[:3]
        MAJOR = pypy_version_info[0]
        MINOR = pypy_version_info[1]
        PATCH = pypy_version_info[2]
        if not isinstance(period, float):
            raise ValueError("You need to pass a float as an argument")
        if warn and pypy_version_info < (4, 1, 0):
            raise Exception("PyPy <4.1 have various kinds of bugs, pass warn=False if you know what you're doing")
        if warn and memory:
            print("Memory profiling is currently unsupported for PyPy. Running without memory statistics.")
        if warn and lines:
            print('Line profiling is currently unsupported for PyPy. Running without lines statistics.\n')
        native = _is_native_enabled(native)
        #
        if (MAJOR, MINOR, PATCH) >= (5, 9, 0):
            _vmprof.enable(fileno, period, memory, lines, native, real_time)
            return
        if real_time:
            raise ValueError('real_time=True requires PyPy >= 5.9')
        if MAJOR >= 5 and MINOR >= 8 and PATCH >= 0:
            _vmprof.enable(fileno, period, memory, lines, native)
            return
        _vmprof.enable(fileno, period)
else:
    # CPYTHON
    def enable(fileno, period=DEFAULT_PERIOD, memory=False, lines=False, native=None, real_time=False):
        if not isinstance(period, float):
            raise ValueError("You need to pass a float as an argument")
        native = _is_native_enabled(native)
        _vmprof.enable(fileno, period, memory, lines, native, real_time)

    def sample_stack_now(skip=0):
        """ Helper utility mostly for tests, this is considered
            private API.

            It will return a list of stack frames the python program currently
            walked.
        """
        stackframes = _vmprof.sample_stack_now(skip)
        assert isinstance(stackframes, list)
        return stackframes

    def resolve_addr(addr):
        """ Private API, returns the symbol name of the given address.
            Only considers linking symbols found by dladdr.
        """
        return _vmprof.resolve_addr(addr)

def insert_real_time_thread():
    """ Inserts a thread into the list of threads to be sampled in real time mode.
        When enabling real time mode, the caller thread is inserted automatically.
        Returns the number of registered threads, or -1 if we can't insert thread.
    """
    return _vmprof.insert_real_time_thread()

def remove_real_time_thread():
    """ Removes a thread from the list of threads to be sampled in real time mode.
        When disabling in real time mode, *all* threads are removed automatically.
        Returns the number of registered threads, or -1 if we can't remove thread.
    """
    return _vmprof.remove_real_time_thread()


def is_enabled():
    """ Indicates if vmprof has already been enabled for this process.
        Returns True or False. None is returned if the state is unknown.
    """
    if hasattr(_vmprof, 'is_enabled'):
        return _vmprof.is_enabled()
    raise NotImplementedError("is_enabled is not implemented on this platform")

def get_profile_path():
    """ Returns the absolute path for the file that is currently open.
        None is returned if the backend implementation does not implement that function,
        or profiling is not enabled.
    """
    if hasattr(_vmprof, 'get_profile_path'):
        return _vmprof.get_profile_path()
    raise NotImplementedError("get_profile_path not implemented on this platform")
